#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <helper_cuda.h>

#define N 100
#define NB_STEPS 5000

/* uncomment this line to activate the debug mode */
//#define DEBUG 1
#ifdef DEBUG
#define DEBUG_PRINTF printf
#else
#define DEBUG_PRINTF //printf
#endif

float *cur_step_h;
float *next_step_h;

float *cur_step_d;
float *next_step_d;

#define ACCESS_GENERIC(array, i, j, pitch)  array[((i)*(pitch))+(j)]

#define ACCESS(array, i, j)  ACCESS_GENERIC(array, i, j, N)

/* initialize matrixes */
void init() {
  cur_step_h = (float*) calloc(N * N, sizeof(float));
  next_step_h = (float*) calloc(N * N, sizeof(float));

  srand(time(NULL));
  int posx = rand()%N;
  int posy = rand()%N;


  ACCESS(cur_step_h, posx, posy) = N*10000;

  /* allocate matrixes on the GPU */
  cudaMalloc( (void**)&cur_step_d, (size_t)(N*N*sizeof(float)));
  cudaMalloc( (void**)&next_step_d, (size_t)(N*N*sizeof(float)));

  /* copy data to the GPU */
  checkCudaErrors(cudaMemcpy(cur_step_d, cur_step_h, N*N*sizeof(float), cudaMemcpyHostToDevice));
  checkCudaErrors(cudaMemcpy(next_step_d, next_step_h, N*N*sizeof(float), cudaMemcpyHostToDevice));
}

/* dump the matrix in f */
void print_matrix(FILE* f,float* matrix)
{
  int i, j;
int zero=0;
  for(i=1; i<N-1; i++) {
zero = 0;
    for(j=1; j<N-1; j++) {
      fprintf(f, "%.2f  ", ACCESS(matrix, i, j));
      if(ACCESS(matrix, i, j) > 0.001)
      			zero++;
    }
    fprintf(f, "\n");
    if(!zero)
    fprintf(stderr, "line %d is filled with zeroes\n", i);
  }
}

/* dump all the submatrixes in a file */
void print_all(FILE* f)
{
  print_matrix(f, cur_step_h);
}


/***    Le kernel   ***/
__global__ void compute_gpu(float *cur_step, float*next_step)
{

  /* This kernel is absolutely not optimized. We just what to make it work for now. Performance doesn't matter. */
  int my_id = blockIdx.x * blockDim.x + threadIdx.x;
  int x = 1+(my_id%(N-2));
  int y = 1+(my_id/(N-2));
#if 1
     ACCESS_GENERIC(next_step, y, x, N) =
      (ACCESS_GENERIC(cur_step, y  , x, N) +
       ACCESS_GENERIC(cur_step, y  , x+1, N) +
       ACCESS_GENERIC(cur_step, y  , x-1, N) +
       ACCESS_GENERIC(cur_step, y+1, x, N) +
       ACCESS_GENERIC(cur_step, y-1, x, N)
  	)/5;
#else
     ACCESS_GENERIC(next_step, y, x, N) = 1;
#endif
}


void compute()
{
  /* offload the computation on the GPU */
  compute_gpu<<<N-2,N-2>>>(cur_step_d, next_step_d);

  DEBUG_PRINTF("compute done\n");

  float * tmp_d = cur_step_d;
  cur_step_d = next_step_d;
  next_step_d = tmp_d;

  float * tmp_h = cur_step_h;
  cur_step_h = next_step_h;
  next_step_h = tmp_h;
}


int main(int argc, char**argv) {

  struct timeval t1, t2;

  init();

  printf("Start computing (%d steps)\n", NB_STEPS);
  gettimeofday(&t1, NULL);
  int i;

  for(i=0; i< NB_STEPS; i++) {
    printf("STEP %d...\n", i);
    compute();
  }

  /* wait untill the computation is over */
  checkCudaErrors(cudaDeviceSynchronize( ));
  checkCudaErrors(cudaMemcpy(cur_step_h, cur_step_d, N*N*sizeof(float), cudaMemcpyDeviceToHost));

  gettimeofday(&t2, NULL);

  float total_time = ((t2.tv_sec-t1.tv_sec)*1e6 + (t2.tv_usec - t1.tv_usec))/1e6;
  printf("%d steps in %lf sec (%lf sec/step)\n",
  	   NB_STEPS, total_time, total_time/NB_STEPS);

  if(argc>1){
    printf("dumping the result data in %s\n", argv[1]);
    FILE *f = fopen(argv[1], "w");
    print_all(f);
  }

  return 0;
}
